program CD_player;
{$G+,I-,R-,S-}
uses    CRT;
Const   InStandBy   : Boolean = FALSE;

        Status_Bits : record
                          Error : byte;
                          Busy  : byte;
                          Done  : byte;
                      end = (Error:$80; Busy:$2; Done:$1);

        IOCTL_INPUT  = 003;     { Read                        }
        IOCTL_OUTPUT = 012;     { Write                       }
        SEEK_CMD     = 131;     { Seek da head ! not tested ! }
        SEEK_LEN     = 024;
        PLAY_CMD     = 132;     { Playing                     }
        PLAY_LEN     = 022;
        STOP_CMD     = 133;     { 1st STOP (PAUSE), 2nd STOP  }
        STOP_LEN     = 005;
        RESUME_CMD   = 136;     { Continue play after 1 stop  }
        RESUME_LEN   = 005;

        EJECT_CMD    = 000;     { Eject Disk                  }
        EJECT_LEN    = 001;
        U_LOCK_CMD   = 001;     { Lock/Unlock Door            }
        U_LOCK_LEN   = 002;
        RESET_CMD    = 002;     { Reset Drive                 }
        RESET_LEN    = 001;
        AUDIO_CMD    = 004;     { Audio Channel Control       }
        AUDIO_LEN    = 009;
        WRITE_CMD    = 004;     { Write Device Control String }
        WRITE_LEN    = 000;
        CLOSE_CMD    = 005;     { Close Tray                  }
        CLOSE_LEN    = 001;

        LOCATE_CMD   = 001;     { Location of the head        }
        LOCATE_LEN   = 006;
        STATUS_CMD   = 006;     { Device Status               }
        STATUS_LEN   = 005;
        SETVOL_CMD   = 003;     { Return Volume Size          }
        SETVOL_LEN   = 009;
        DISKINF_CMD  = 010;     { Audio Disk Info             }
        DISKINF_LEN  = 007;
        TRACKINF_CMD = 011;     { Audio Track Info            }
        TRACKINF_LEN = 007;
        Q_INFO_CMD   = 012;     { Audio Q-Channel Info        }
        Q_INFO_LEN   = 011;
        AUD_STAT_CMD = 015;     { Audio Status (door closed?) }
        AUD_STAT_LEN = 011;

Type
        str50            = string[50];
        str05            = string[05];
        T_Request_header = object
                              Length    : byte;
                              Subunit   : byte;
                              Command   : byte;
                              Error     : byte;
                              Status    : byte;
                              Reserved  : array [1..8] of byte;
                           end;
        P_Request_header = ^T_Request_header;

        T_Req_data_header= object(T_Request_header)
                              data      : array [1..20] of byte;
                           end;
        P_Req_data_header= ^T_Req_data_header;

        T_IOCtrl_Header  = object(T_Request_header)
                              MediaDesc : byte;
                              Transfer  : pointer;
                              NoOfBytes : word;
                              StartSect : word;
                              vol_ID    : longint;
                           end;
        P_IOCtrl_Header  = ^T_IOCtrl_Header;

        T_Transfer       = object
                              command   : byte;
                              data      : array [1..20] of byte;
                           end;
        P_Transfer       = ^T_Transfer;

        T_CDROMlist      = array[0..25] of boolean;

        T_Disk_info      = record
                                 Lowest_track   : byte;
                                 Highest_track  : byte;
                                 Last_pos       : longint;
                                 Min_CD         : byte;
                                 Sec_CD         : byte;
                                 Frame_CD       : byte;
                           end;

        T_Track_info     = record
                                 Current_track  : byte;
                                 Minutes        : byte;
                                 Seconds        : byte;
                                 Frame          : byte;
                           end;

        T_AudioChannels  = record
                                 Input_0        : byte;
                                 Volume_0       : byte;
                                 Input_1        : byte;
                                 Volume_1       : byte;
                                 Input_2        : byte;
                                 Volume_2       : byte;
                                 Input_3        : byte;
                                 Volume_3       : byte;
                           end;
        AudioStatus      = (playing, paused, stop, unknown);

var     CDROMlist        : T_CDROMlist;
        CDROM_info       : T_Disk_info;
        Track_info       : T_Track_info;
        AudioChannels    : T_AudioChannels;


procedure ShutDaCursor; assembler;
    asm
        mov   ah,01h
        mov   cx,2020h
        int   10h
    end;

procedure ShowDaCursor; assembler;
    asm
        mov   ah,01h
        mov   cx,0708h
        int   10h
    end;


function GetErrorMSG(error: byte): str50;  { CD-ROM errors that may occur }
var  strng : string[50];
begin
     case error of
      $00 :  strng:='Write-protect violation';
      $01 :  strng:='Unknown unit';
      $02 :  strng:='Drive not ready';
      $03 :  strng:='Unknown command';
      $04 :  strng:='CRC error';
      $05 :  strng:='Bad drive request structure length';
      $06 :  strng:='Seek error';
      $07 :  strng:='Unknown media';
      $08 :  strng:='Sector not found';
      $09 :  strng:='Printer out of paper (!?)';
      $0a :  strng:='Write fault';
      $0b :  strng:='Read fault';
      $0c :  strng:='General failure';
      $0d :  strng:='Reserved';
      $0e :  strng:='Media unavailable ';
      $0f :  strng:='Invalid disk change';
        else strng:='Imposibile error-code, programmer fault';
     end;
     GetErrorMSG:=strng;
end;

procedure WriteErrorMSG(stng: str50);
begin
     writeln('ERR: ',stng);
     ShowDaCursor;
     Halt(1);
end;

{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
function Add0(w : Word) : Str50;    { Make val2str, leading zero }
var  s : String[2];
begin
     Str(w:0,s);
     if Length(s) = 1 then s := '0' + s;
     Add0 := s;
end;


function AddStatus2(stat: audiostatus): str05;
const
     pos  : byte = 0;
     play : array [1..4] of str05 = ('(|)','(/)','(-)','(\)');
     sto_ : array [1..4] of str05 = ('()','( )','()','( )');
     paus : array [1..4] of str05 = ('()','( )','()','( )');
     unkn : array [1..4] of str05 = ('(?)','( )','(?)','( )');
begin
     inc(pos);
     if pos=5 then pos:=1;
     case stat of
       Playing : AddStatus2:=play[pos];
       Paused  : AddStatus2:=paus[pos];
       Stop    : AddStatus2:=sto_[pos];
       Unknown : AddStatus2:=unkn[pos];
     end;
end;



function AddStatus(stat: audiostatus): str05;
const
     pos  : byte = 0;
     play : array [1..4] of str05 = ('>   ',' >  ','  > ','   >');
     sto_ : array [1..4] of str05 = ('*TOP','S*OP','ST*P','STO*');
     paus : array [1..4] of str05 = ('>><<',' >< ','    ','>><<');
     unkn : array [1..4] of str05 = (' ?? ','?  ?','    ',' ?? ');
begin
     inc(pos);
     if pos=5 then pos:=1;
     case stat of
      playing: AddStatus:=play[pos];
      stop:    AddStatus:=sto_[pos];
      paused:  AddStatus:=paus[pos];
      unknown: AddStatus:=unkn[pos];
     end;
end;
{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}



{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
procedure WriteAbout;
begin
     Writeln;
     Writeln('CD-PLAY.EXE drive:');
     Writeln('drive: - Optional parameter discribes which');
     Writeln('         CD-ROMed drive must play audio disk');
     Writeln('         drive: = A:, B:, C:, D: and so on.');
     Writeln;
     Writeln('- Thanx to Radek Gedzba whos borrowing me CD-ROM -');
     Writeln('  // But now I have my own Quad Speed CD-ROM! //');
     Writeln;
     Writeln('My E-mail Box: p_lark@hotmail');
     Writeln;
     Halt;
end;


{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
function cdrom_check: boolean;   { True= Installed }
assembler;
asm
        mov   ax,1100h
        push  0DADAh
        int   2Fh
        pop   bx
        or    al,al
        jz    @not_installed
        cmp   al,01h
        jz    @not_installed
        cmp   bx,0DADAh
        jz    @not_installed

        jmp   @installed

@not_installed:
        xor   ax,ax
        jmp   @exiting
@installed:
        mov   ax,1
@exiting:
end;


function CDROMver: word;    { Get MSCDEX version }
assembler;
asm
        mov   ax,150Ch
        int   2Fh
        mov   ax,bx
end;


procedure SendHeaderRequest(drv: byte; request: P_Request_Header);
var  oft, sgm: word;
begin
     oft:=ofs(request^);
     sgm:=seg(request^);
     asm
        push   es
        mov    ax,1510h
        xor    cx,cx
        mov    cl,drv
        mov    bx,oft
        mov    dx,sgm
        mov    es,dx
        int    2Fh
        pop    es
     end;
end;

procedure SendIOCtrlRequest(drv: byte; IOCtrl   : P_IOCtrl_Header;
                                       Transfer : P_Transfer);
begin
     IOCtrl^.Transfer:=Transfer;
     SendHeaderRequest(drv,IOCtrl);
end;

function IsCDROMed(drv: byte): boolean;   { Check if Drive is CD-ROMed }
assembler;
asm
        push  ds
        push  es
        push  di
        mov   ax,150Bh
        xor   cx,cx
        mov   cl,drv
        int   2Fh
        pop   di
        pop   es
        pop   ds
        cmp   bx,0ADADh
        jnz   @not_mscdex
        or    ax,ax
        jz    @not_cdromed
        jmp   @cdromed

@not_mscdex:
@not_cdromed:
        xor   ax,ax
        jmp   @exiting
@cdromed:
        mov   ax,1
@exiting:
end;

procedure CDROMedDrives(var CDROMed: T_CDROMlist); { Makes a list of CD-ROMed }
var  i: byte;                                      { drives                   }
begin
     for i:=0 to 25 do
     if IsCDROMed(i) Then CDROMed[i]:=True
                     Else CDROMed[i]:=False;
end;

function GetCDROMStatus(drv: byte): longint; { Door open/closed itd }
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
     long     : longint;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_INPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=STATUS_LEN;
     Transfer.command:=STATUS_CMD;

     SendIOCtrlRequest(drv,@IOCtrl,@Transfer);
     if IOCtrl.Status and Status_Bits.Done<>Status_Bits.Done then writeln('FUCK');

     Move(Transfer.data[1],long,4);
     GetCDROMStatus:=long;
end;
{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}


function Red2HSG(red: longint): longint;       { Red Book spec 2 High Sierra }
var  long : longint;
begin
     long:=((red shr 16) and $FF) * 60 * 75;   { Min }
     long:=long + ((red shr 8) and $FF) * 75;  { Sec }
     long:=long + red and $FF - 150;           { Frame }
     Red2HSG:=long;
end;




{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
function GetTrackInfo(drive,track: byte; transfer: P_Transfer): byte; { About track info }
var  IOCtrl   : T_IOCtrl_Header;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer^,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_INPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=TRACKINF_LEN;
     Transfer^.command:=TRACKINF_CMD;
     Transfer^.data[1]:=track;

     SendIOCtrlRequest(drive,@IOCtrl,Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then GetTrackInfo:=0
                                                            else GetTrackInfo:=IOCtrl.Error;
end;


function GetDiskInfo(drive: byte; transfer: P_Transfer): byte; { About disk info }
var  IOCtrl   : T_IOCtrl_Header;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer^,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_INPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=DISKINF_LEN;
     Transfer^.command:=DISKINF_CMD;

     SendIOCtrlRequest(drive,@IOCtrl,Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then GetDiskInfo:=0
                                                            else GetDiskInfo:=IOCtrl.Error;
end;

function GetQchannel(drive: byte; Transfer: P_Transfer): byte;
var  IOCtrl   : T_IOCtrl_Header;             { A lot useful infos }
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer^,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_INPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=Q_INFO_LEN;
     Transfer^.command:=Q_INFO_CMD;

     SendIOCtrlRequest(drive,@IOCtrl,Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then GetQchannel:=0
                                                            else GetQchannel:=IOCtrl.Error;
end;


procedure GetCDVol(drive: byte; var AudioChannels: T_AudioChannels);
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
     long     : longint;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_INPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=AUDIO_LEN;
     Transfer.command:=AUDIO_CMD;

     SendIOCtrlRequest(drive,@IOCtrl,@Transfer);

     AudioChannels.Input_0:=transfer.data[1];
     AudioChannels.Volume_0:=transfer.data[2];
     AudioChannels.Input_1:=transfer.data[3];
     AudioChannels.Volume_1:=transfer.data[4];
     AudioChannels.Input_2:=transfer.data[5];
     AudioChannels.Volume_2:=transfer.data[6];
     AudioChannels.Input_3:=transfer.data[7];
     AudioChannels.Volume_3:=transfer.data[8];
end;

{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
procedure SetCDVol(drive: byte; AudioChannels: T_AudioChannels);
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
     long     : longint;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_OUTPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=SETVOL_LEN;
     Transfer.command:=SETVOL_CMD;

     transfer.data[1]:=AudioChannels.Input_0;
     transfer.data[2]:=AudioChannels.Volume_0;
     transfer.data[3]:=AudioChannels.Input_1;
     transfer.data[4]:=AudioChannels.Volume_1;
     transfer.data[5]:=AudioChannels.Input_2;
     transfer.data[6]:=AudioChannels.Volume_2;
     transfer.data[7]:=AudioChannels.Input_3;
     transfer.data[8]:=AudioChannels.Volume_3;

     SendIOCtrlRequest(drive,@IOCtrl,@Transfer);
end;


{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
Function CDAudioStatus(drive: byte): AudioStatus;
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
     status   : word;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_INPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=AUD_STAT_LEN;
     Transfer.command:=AUD_STAT_CMD;

     SendIOCtrlRequest(drive,@IOCtrl,@Transfer);

     if IOCtrl.Status and Status_Bits.Error<>0 Then
        begin
             CDAudioStatus:=Unknown;
             exit;
        end;

     Move(Transfer.data[1],status,2);

     if (IOCtrl.Status and Status_Bits.Busy <> 0) and
        (status and 1=0) { Not pausing } Then
        begin
             CDAudioStatus:=Playing;
             exit;
        end;

     if status and 1=1 then CDAudioStatus:=Paused
                       else CDAudioStatus:=Stop;
end;



{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
function GetInfos(drive: byte; var CDROMinfo: T_Disk_info;
                               var TRACKinfo: T_Track_info): byte;
                               { Times, Max tracks itd }
var  Transfer : T_Transfer;
begin
     GetDiskInfo(drive,@transfer);

     CDROMinfo.lowest_track:=Transfer.data[1];
     CDROMinfo.highest_track:=Transfer.data[2];
     Move(Transfer.data[3],CDROMinfo.last_pos,4);

     GetQchannel(drive,@transfer);

     CDROMinfo.Min_CD:=Transfer.data[4];
     CDROMinfo.Sec_CD:=Transfer.data[5];
     CDROMinfo.Frame_CD:=Transfer.data[6];

     TrackInfo.current_track:=Transfer.data[2];

     TrackInfo.Minutes:=Transfer.data[8];
     TrackInfo.Seconds:=Transfer.data[9];
     TrackInfo.Frame:=Transfer.data[10];
end;


{%%%%%%%%%%%%%%%%%%%%%%%%%%%%   A U D I O   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
function EjectDisk(drive: byte): byte;
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_OUTPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=EJECT_LEN;
     Transfer.command:=EJECT_CMD;

     SendIOCtrlRequest(drive,@IOCtrl,@Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then EjectDisk:=0
                                                            else EjectDisk:=IOCtrl.Error;
end;

function CloseDisk(drive: byte): byte;
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_OUTPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=CLOSE_LEN;
     Transfer.command:=CLOSE_CMD;

     SendIOCtrlRequest(drive,@IOCtrl,@Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then CloseDisk:=0
                                                            else CloseDisk:=IOCtrl.Error;
end;

function ULockDisk(drive,uLock: byte): byte;  { Lock/UnLock disk }
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_OUTPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=U_LOCK_LEN;
     Transfer.command:=U_LOCK_CMD;
     Transfer.data[1]:=uLock;

     SendIOCtrlRequest(drive,@IOCtrl,@Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then ULockDisk:=0
                                                            else ULockDisk:=IOCtrl.Error;
end;

function ResetCDROM(drive: byte): byte;
var  IOCtrl   : T_IOCtrl_Header;
     Transfer : T_Transfer;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_OUTPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=RESET_LEN;
     Transfer.command:=RESET_CMD;

     SendIOCtrlRequest(drive,@IOCtrl,@Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then ResetCDROM:=0
                                                            else ResetCDROM:=IOCtrl.Error;
end;

function StopAudio(drive: byte): byte;
var  Request : T_Request_Header;
begin
     FillChar(Request,sizeof(T_Request_Header),0);
     Request.command:=STOP_CMD;
     Request.length:=STOP_LEN;
     SendHeaderRequest(drive,@Request);

     if Request.Status and Status_Bits.Done=Status_Bits.Done then StopAudio:=0
                                                             else StopAudio:=Request.Error;
end;

function ResumeAudio(drive: byte): byte;
var  Request : T_Request_Header;
begin
     FillChar(Request,sizeof(T_Request_Header),0);
     Request.command:=RESUME_CMD;
     Request.length:=RESUME_LEN;
     SendHeaderRequest(drive,@Request);
     if Request.Status and Status_Bits.Done=Status_Bits.Done then ResumeAudio:=0
                                                             else ResumeAudio:=Request.Error;
end;

function HeadPosition(drive,track: byte; transfer: P_Transfer): byte; { Head location }
var  IOCtrl   : T_IOCtrl_Header;
begin
     FillChar(IOCtrl,sizeof(T_IOCtrl_Header),0);
     FillChar(Transfer^,sizeof(T_Transfer),0);
     IOCtrl.command:=IOCTL_INPUT;
     IOCtrl.length:=sizeof(T_IOCtrl_Header);
     IOCtrl.NoOfBytes:=LOCATE_LEN;
     Transfer^.command:=LOCATE_CMD;
     Transfer^.data[1]:=0;

     SendIOCtrlRequest(drive,@IOCtrl,Transfer);

     if IOCtrl.Status and Status_Bits.Done=Status_Bits.Done then HeadPosition:=0
                                                            else HeadPosition:=IOCtrl.Error;
end;



function SeekHead(drive,track: byte; seek,last_pos: longint): byte;
var  Request  : T_Req_data_Header;
     Transfer : T_Transfer;
     long     : longint;
begin
     FillChar(Request,sizeof(T_Req_data_Header),0);
     Request.command:=PLAY_CMD;
     Request.length:=PLAY_LEN;

     StopAudio(Drive);
     HeadPosition(drive, track, @transfer);
(*     Request.data[1]:=0;             { uncheck by driver if HSG or RED } *)
     Move(transfer.data[2], long, 4);  { Point returned as HSG }
     if long+seek>0 then long:=long+seek
                    else long:=long;

     Move(long,Request.data[2],4);     { From position }
     long:=Red2HSG(last_pos)-long;
     Move(long,Request.data[6],4);     { No of sec to play }


     SendHeaderRequest(drive,@Request);

     if Request.Status and Status_Bits.Done=Status_Bits.Done then SeekHead:=0
                                                             else SeekHead:=Request.Error;
end;

function PlayAudio(drive, track: byte; last_pos: longint): byte;
var  Request  : T_Req_data_Header;
     Transfer : T_Transfer;
     long     : longint;
{
 PlayReq   DB   13 dup (0); Request header
           DB   ?         ; Addressing mode
           DD   ?         ; Starting sector number
           DD   ?         ; Number of sectors to read
}
begin
     FillChar(Request,sizeof(T_Req_data_Header),0);
     Request.command:=PLAY_CMD;
     Request.length:=PLAY_LEN;

     GetTrackInfo(drive, track, @transfer);
     Request.data[1]:=0;               { "Red Book" addressing mode }
     Move(transfer.data[2], long, 4);
     long:=Red2HSG(long);
     Move(long,Request.data[2],4);     { From position }
     long:=Red2HSG(last_pos)-long;
     Move(long,Request.data[6],4);     { No of sec to play }

     SendHeaderRequest(drive,@Request);

     if Request.Status and Status_Bits.Done=Status_Bits.Done then PlayAudio:=0
                                                             else PlayAudio:=Request.Error;
end;
{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}


procedure WriteStatus(drive: byte);       { CD-ROM driver status }
var  long: longint;
begin
     Writeln('Device status:');
     long:=GetCDROMStatus(drive);
     if long and 1 = 0 then writeln('      Door closed')
                       else writeln('      Door open');
     if long and 2 = 0 then writeln('      Door locked')
                       else writeln('      Door unlocked');
     if long and 4 = 0 then writeln('      Supports only cooked reading')
                       else writeln('      Supports cooked and raw reading');
     if long and 8 = 0 then writeln('      Read only')
                       else writeln('      Read & write');
     if long and 16= 0 then writeln('      No interleaving')
                       else writeln('      Supports interleaving');
{     if long and 32= 0 then writeln('Reserved')
                       else writeln('Reserved');}
     if long and 64= 0 then writeln('      No prefetching')
                       else writeln('      Supports prefetching requests');
     if long and 128=0 then writeln('      No audio channel manipulation')
                       else writeln('      Supports audio channel manipulation');
     if long and 256=0 then writeln('      Supports HSG addressing mode')
                       else writeln('      Supports HSG and Red Book addressing modes');
end;


{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
procedure PrintWhishes;
begin
     Writeln;
     Writeln('What''s your wish ?');
     Writeln('     --- [T]urn to Stand-by  [I]nfo  [ESC] Exit to OS --- ');
     Writeln('[P]lay  [S]top  [R]esume  r[E]set  [O]pen  [C]lose  [U]n/[L]ock');
end;

{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
procedure CleanDaBuffer;
begin
     while keypressed do readkey;
end;


{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
function GetDriveNo: byte;
var  DriveNo,i : byte;
     strng     : str50;
begin
     if paramcount=0 then       { On which drive play (user passed drive) }
        begin
             DriveNo:=byte(-1);
             For i:=0 to 25 do  { Find first available CD-ROMed drive }
                 if CDROMlist[i] then begin DriveNo:=i; break; end;
             if DriveNo=byte(-1) Then WriteErrorMSG('None of CD-ROMed drives found');
             Writeln('Working on first available CD-ROMed drive ',Chr(DriveNo+Ord('A')),':');
        end
           else
        begin
             strng:=paramstr(1);
             DriveNo:=Ord(UpCase(strng[1]))-Ord('A');
             If NOT CDROMlist[DriveNo] then WriteErrorMSG('Not CD-ROMed drive.');
             Writeln('Working on CD-ROMed drive ',Chr(DriveNo+Ord('A')),':');
        end;
     GetDriveNo:=DriveNo;
end;

{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
procedure TurnToStandBy;
var  al_,ah_ : byte;
     count,i : byte;
begin
     for count:=5 downto 0 do
     begin
          write(#13,'Turning to stand-by... and count-down [',count:2,'] <!SPACE! - to wake>');
          if keypressed then
             begin
                  readkey;
                  write(#13);
                  clreol;
                  writeln(#13,'Turning to Stand-by failed.');
                  PrintWhishes;
                  exit;
             end;
          for i:=1 to 10 do
              begin
                   Delay(100);
                   if keypressed then break;
              end;
     end;
     write(#13);
     clreol;
     port[$1F6]:=$0;       { Drive: 4bit=0 C: }     { Rejestr nepdu/gowicy }
     port[$1F7]:=$E0;      { Command: E0h     }     { Rejestr stanu/rozkazu  }
     port[$1F6]:=$10;      { Drive: 4bit=1 D: }     { Rejestr nepdu/gowicy }
     port[$1F7]:=$E0;      { Command: E0h     }     { Rejestr stanu/rozkazu  }

     asm                   { Set display power state }
        mov    ax,4f10h
        mov    bl,01h
        mov    bh,1
        int    10h
        mov    al_,al
        mov    ah_,ah
     end;
     if al_=$4f then
        writeln('Machine is now in stand-by mode (PM supported).')
     else
        writeln('Machine is now in stand-by mode...I think (PM NOT supported).');

     InStandBy:=TRUE;
     PrintWhishes;
end;

procedure WakeFromStandBy;
var  al_, ah_ : byte;
begin
     if InStandBy Then Writeln('Awakening...')
                  else clreol;
     asm                   { Set display power state }
        mov    ax,4f10h
        mov    bl,01h
        mov    bh,0
        int    10h
        mov    al_,al
        mov    ah_,ah
     end;
     InStandBy:=False;
     PrintWhishes;
end;


{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
procedure Exit2OS;
begin
     ShowDaCursor;
     clreol;
     gotoxy(25,wherey);
     Writeln(' THE END ');
     Halt;
end;

{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
function KeyboardHandler(DriveNo,Track: byte): byte;
var  Error : byte;
     strng : str50;
     ch    : char;
     shit  : integer;
begin
     Error:=0;
     KeyboardHandler:=0;
     case UpCase(readkey) of
     'O' : Error:=EjectDisk(DriveNo);
     'C' : Error:=CloseDisk(DriveNo);
     'L' : Error:=ULockDisk(DriveNo,1);
     'U' : Error:=ULockDisk(DriveNo,0);
     'E' : Error:=ResetCDROM(DriveNo);
     'S' : Error:=StopAudio(DriveNo);
     'R' : Error:=ResumeAudio(DriveNo);
     '+' : begin
                GetCDVol(DriveNo,AudioChannels);
                if AudioChannels.Volume_0+10<255 then inc(AudioChannels.Volume_0,10);
                if AudioChannels.Volume_1+10<255 then inc(AudioChannels.Volume_1,10);
                if AudioChannels.Volume_2+10<255 then inc(AudioChannels.Volume_2,10);
                if AudioChannels.Volume_3+10<255 then inc(AudioChannels.Volume_3,10);
                SetCDVol(DriveNo,AudioChannels);
           end;
     '-' : begin
                GetCDVol(DriveNo,AudioChannels);
                if AudioChannels.Volume_0-10>0 then dec(AudioChannels.Volume_0,10);
                if AudioChannels.Volume_1-10>0 then dec(AudioChannels.Volume_1,10);
                if AudioChannels.Volume_2-10>0 then dec(AudioChannels.Volume_2,10);
                if AudioChannels.Volume_3-10>0 then dec(AudioChannels.Volume_3,10);
                SetCDVol(DriveNo,AudioChannels);
           end;
     'T' : TurnToStandBy;
     #32 : WakeFromStandBy;
     'I' : begin WriteStatus(DriveNo); PrintWhishes; end;
     'P',#13: begin                       { Play track... }
                with CDROM_Info do Write('Which track (',Add0(lowest_track),'-',Add0(highest_track),') ? ');
                strng:='';
                ShowDaCursor;
                repeat            { Readln, but not jmp to next line }
                      ch:=readkey;
                      case ch of
                       #13 : break;
                       '0'..'9': begin
                                      write(ch);
                                      strng:=strng+ch;
                                 end;
                       #08 : if length(strng)>0 then
                                begin
                                     delete(strng,length(strng),1);
                                     write(#8,' ',#8);
                                end;
                       #27 : begin
                                  strng:='';
                                  break;
                             end;
                      end;
                until false;
                ShutDaCursor;
                if strng='' then exit;
                val(strng,Track,Shit);
                if (Track<CDROM_Info.lowest_track) or (Track>CDROM_Info.Highest_track) then
                   Track:=CDROM_Info.lowest_track;
                StopAudio(DriveNo);
                Error:=PlayAudio(DriveNo,Track,CDROM_Info.last_pos);
           end;
      #27: if InStandBy Then WakeFromStandBy  { Wake PC }
                        Else Exit2OS;         { Exiting.... }
      #00: begin
                case UpCase(ReadKey) of
      { Left  } #75: Error:=SeekHead(DriveNo,Track,-5000,CDROM_Info.last_pos);
      { Right } #77: Error:=SeekHead(DriveNo,Track,+5000,CDROM_Info.last_pos);
      { Home  } #71: begin
                          Track:=CDROM_Info.Lowest_track;
                          StopAudio(DriveNo);
                          Error:=PlayAudio(DriveNo,Track,CDROM_Info.last_pos);
                     end;
      {  End  } #79: begin
                          Track:=CDROM_Info.Highest_track;
                          StopAudio(DriveNo);
                          Error:=PlayAudio(DriveNo,Track,CDROM_Info.last_pos);
                     end;
      {  Up   } #72: begin
                          inc(Track);
                          if (Track<CDROM_Info.lowest_track) or (Track>CDROM_Info.Highest_track) then
                          Track:=CDROM_Info.lowest_track;
                          StopAudio(DriveNo);
                          Error:=PlayAudio(DriveNo,Track,CDROM_Info.last_pos);
                     end;
      { Down  } #80: begin
                          dec(Track);
                          if (Track<CDROM_Info.lowest_track) or (Track>CDROM_Info.Highest_track) then
                          Track:=CDROM_Info.Highest_track;
                          StopAudio(DriveNo);
                          Error:=PlayAudio(DriveNo,Track,CDROM_Info.last_pos);
                     end;
                end;
           end;
      else begin
                CleanDaBuffer;
                Write('Careful what you wish ! -- press something --',#7);
                readkey;
                CleanDaBuffer;
                Exit;
           end;
     end;
     KeyboardHandler:=Error;
end;


{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
{                                   M A I N                                   }
{%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%}
var  DriveNo, Track, Error, i : byte;
begin
     Writeln('CD player  -- by Piotrek Skowronek -- 01 Jul 1997 -- [FreeSHIT]');
     if paramstr(1)='/?' then WriteAbout;
     Writeln;
     Write('MSCDEX driver: ');
     if NOT cdrom_check then WriteErrorMSG('Not Installed.'); { MSCDEX not installed }

     Writeln('installed, ver ',CDROMver shr 8,'.',CDROMver and $FF:2); { Ver }
     CDROMedDrives(CDROMlist);
     Write('List of CD-ROMed drives: [');  { Print list of CD-ROMed drives }
     For i:=0 to 25 do
         if CDROMlist[i] then write(chr(i+Ord('A')),': ');
     Writeln(']');

     DriveNo:=GetDriveNo;


     PrintWhishes;
     ShutDaCursor;
REPEAT
     repeat
           GetInfos(DriveNo, CDROM_info, Track_Info);
           Track:=Track_Info.current_track;
           Repeat Until port[$3da] and 8=8;      { Vertical return }
           with Track_Info do
           with CDROM_Info do
           Write(#13,'[Tracks ',Add0(lowest_track),'-',Add0(highest_track),
                     '  ',Add0(last_pos shr 16 and $FF),':',Add0(last_pos shr 8 and $FF),']',
                     '  Track ',track shr 4,track and $F,
                     '  Time: total ',Add0(Minutes),':',Add0(Seconds),{'.',Add0(Frame),}
                     ', track ',Add0(Min_CD),':',Add0(Sec_CD),{'.',Add0(frame_CD),}'    ');
{           write(AddStatus(CDAudioStatus(DriveNo)));}
           write(AddStatus2(CDAudioStatus(DriveNo)));
           delay(100);
     until KeyPressed;
           Write(#13);
           ClrEol;
           Write('WAIT...',#13);
           Error:=KeyboardHandler(DriveNo,Track);
           if Error<>0 Then Writeln('Fuck, I knew it: ',GetErrorMSG(Error),'.');

UNTIL FALSE;
end.